import logging
import math
import os
import random
import time
from typing import List, Union, Tuple

import numpy as np
import pygame
import typing
from OpenGL.GL import glClear, GL_COLOR_BUFFER_BIT, glClearColor, glBindVertexArray, \
    glBindBuffer, GL_ARRAY_BUFFER, glCreateShader, \
    GL_VERTEX_SHADER, glShaderSource, glCompileShader, glCreateProgram, glAttachShader, glLinkProgram, glUseProgram, \
    GL_FRAGMENT_SHADER, glPixelStorei, GL_UNSIGNED_BYTE, \
    GL_RGBA, glFlush, glReadPixels, GL_PACK_ALIGNMENT, glEnable, GL_BLEND, glBlendFunc, GL_SRC_ALPHA, \
    GL_ONE_MINUS_SRC_ALPHA, glGetUniformLocation, glUniform4f, glUniform2f, glDepthRange, GL_DEPTH_TEST, GL_DEPTH_BUFFER_BIT, glDepthFunc, GL_LESS, \
    glClearDepth, GL_GREATER, glAlphaFunc, GL_ALPHA_TEST, GL_DEPTH_COMPONENT, GL_FLOAT, glUniform1f
from pygame.time import Clock

from gui import GUIFacade
from gui.input import Keyboard, Mouse, Event
from gui.layer import LayerGroup, Layer, GridLayer
from gui.impl.pygame import PygameKeyboard, PygameMouse, PygameKeyEvent
from state.Region import GROUND_ARRAY_INDEX, TREE_ARRAY_INDEX, WATER_ARRAY_INDEX
from . import ZBuffer
from .layer.OpenGLLayerGroup import OpenGLLayerGroup


MAX_LAYER_COUNT = 256


class OpenGLGUIFacade(GUIFacade):

    def __init__(self):
        # Main properties
        self.__screenWidth = 0  # type: int
        self.__screenHeight = 0  # type: int
        self.__screenPixelWidth = 1  # type: float
        self.__screenPixelHeight = 1  # type: float
        self.__clock = None  # type: Clock
        self.__closingRequested = False  # type: bool

        # Shader properties
        self.__shaderProgramId = -1  # type: int

        # Layer groups
        self.__layerGroups = []  # type: List[Union[None,OpenGLLayerGroup]]

        # Inputs
        self.__events = []  # type: List[Event]
        self.__keyboard = PygameKeyboard()
        self.__mouse = PygameMouse()

    @property
    def screenWidth(self) -> int:
        return self.__screenWidth

    @property
    def screenHeight(self) -> int:
        return self.__screenHeight

    @property
    def screenPixelWidth(self) -> float:
        return self.__screenPixelWidth

    @property
    def screenPixelHeight(self) -> float:
        return self.__screenPixelHeight

    def getTime(self) -> float:
        return pygame.time.get_ticks()

    def createWindow(self, title: str, width: int, height: int, fullScreen: bool = False):
        os.environ['SDL_VIDEO_CENTERED'] = '1'
        pygame.init()
        flags = pygame.DOUBLEBUF | pygame.OPENGL
        if fullScreen:
            flags |= pygame.FULLSCREEN
        pygame.display.set_mode((width, height), flags)
        pygame.display.set_caption(title)
        glClearColor(0.0, 0.0, 0.0, 1.0)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        self.__screenWidth = width
        self.__screenHeight = height
        self.__screenPixelWidth = 2 / float(width)
        self.__screenPixelHeight = 2 / float(height)
        
    # Layer groups 

    def createLayerGroup(self, level: int = -1) -> LayerGroup:
        layerGroup = OpenGLLayerGroup(self)
        if level < 0:
            self.__layerGroups.append(layerGroup)
        elif level < len(self.__layerGroups):
            self.__layerGroups[level] = layerGroup
        else:
            self.__layerGroups.extend([None for l in range(len(self.__layerGroups), level + 1)])
            self.__layerGroups[level] = layerGroup
        return layerGroup
            
    def moveLayerGroup(self, layerGroup: Union[int, LayerGroup], level: int):
        if type(layerGroup) == LayerGroup:
            layerGroup = self.__layerGroups.index(layerGroup)
        if layerGroup < 0 or layerGroup >= len(self.__layerGroups):
            raise ValueError("Invalid level {}".format(layerGroup))
        if level < 0 or level >= len(self.__layerGroups):
            raise ValueError("Invalid level {}".format(level))
        if level != layerGroup:
            self.__layerGroups[level] = self.__layerGroups[layerGroup]
            self.__layerGroups[layerGroup] = None

    def removeLayerGroup(self, layerGroup: Union[int, LayerGroup]):
        if type(layerGroup) == LayerGroup:
            layerGroup = self.__layerGroups.index(layerGroup)
        if layerGroup < 0 or layerGroup >= len(self.__layerGroups):
            raise ValueError("Invalid level {}".format(layerGroup))
        self.__layerGroups[layerGroup] = None

    def layersChanged(self):
        minDepth = ZBuffer.maxDepth - 1
        for layerGroup in self.__layerGroups:
            if layerGroup is None:
                continue
            maxIndex = 0
            for layer in layerGroup:
                if layer is None:
                    continue
                maxIndex += 1
            maxDepth = minDepth
            index = 0
            for layer in layerGroup:
                if layer is None:
                    continue
                minDepth = layer.updateDepths(minDepth, maxDepth, index, maxIndex)
                index += 1

    def __createShaders(self):
        # Release focus on VBO and VAO
        glBindBuffer(GL_ARRAY_BUFFER, 0)
        glBindVertexArray(0)

        # Create vertex shader
        vertexShaderCode = '''#version 330
        layout (location=0) in vec4 vertex;
        layout (location=1) in vec2 inputUV;
                           out vec2 outputUV;
        uniform vec4 translation;
        uniform vec2 uvShift;
        void main() {
            gl_Position = vertex + translation;
            outputUV = inputUV + uvShift;
        }'''
        vertexShaderId = glCreateShader(GL_VERTEX_SHADER)
        glShaderSource(vertexShaderId, vertexShaderCode)
        glCompileShader(vertexShaderId)

        # Create fragment shader
        fragmentShaderCode = '''#version 330
        in vec2 outputUV;
        out vec4 color;
        uniform sampler2D textureColors;
        void main() {
            color = texture(textureColors, outputUV);
        }'''
        fragmentShaderId = glCreateShader(GL_FRAGMENT_SHADER)
        glShaderSource(fragmentShaderId, fragmentShaderCode)
        glCompileShader(fragmentShaderId)

        # Create shader program
        self.__shaderProgramId = glCreateProgram()
        glAttachShader(self.__shaderProgramId, vertexShaderId)
        glAttachShader(self.__shaderProgramId, fragmentShaderId)
        glLinkProgram(self.__shaderProgramId)

    def init(self):
        # Enable Z-Buffer, set far to 1.0 and show lowest values
        glEnable(GL_DEPTH_TEST)
        glClearDepth(1.0)
        glDepthFunc(GL_LESS)
        glDepthRange(0, 1)
        # Only draw pixel with alpha value higher than 0.01
        glAlphaFunc(GL_GREATER, 0.01)
        glEnable(GL_ALPHA_TEST)
        self.__createShaders()
        self.__clock = pygame.time.Clock()

    def getEvents(self) -> List[Event]:
        return self.__events

    def updateInputs(self):
        self.__events.clear()
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.__closingRequested = True
            else:
                guiEvent = self.__keyboard.createEvent(event)
                if guiEvent is not None:
                    self.__events.append(guiEvent)
                    continue
                guiEvent = self.__mouse.createEvent(event)
                if guiEvent is not None:
                    self.__events.append(guiEvent)
                    continue

    def render(self):
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        glUseProgram(self.__shaderProgramId)

        translationShaderVar = glGetUniformLocation(self.__shaderProgramId, "translation")
        uvShiftShaderVar = glGetUniformLocation(self.__shaderProgramId, "uvShift")
        for layerGroup in self.__layerGroups:
            if layerGroup is None:
                continue
            x = layerGroup.translationX + layerGroup.shiftEffectX
            y = layerGroup.translationY + layerGroup.shiftEffectY
            glUniform4f(translationShaderVar, x, y, 0.0, 0.0)
            for layer in layerGroup:
                if layer is None:
                    continue
                glUniform2f(uvShiftShaderVar, layer.uvShiftX, layer.uvShiftY)
                layer.draw()

        glUseProgram(0)

        pygame.display.flip()
        self.__clock.tick(60)

    @property
    def keyboard(self) -> Keyboard:
        return self.__keyboard

    @property
    def mouse(self) -> Mouse:
        return self.__mouse

    @property
    def closingRequested(self) -> bool:
        return self.__closingRequested

    @closingRequested.setter
    def closingRequested(self, b: bool):
        self.__closingRequested = b

    def captureScreenContent(self, x1: int, y1: int, x2: int, y2: int) -> np.ndarray:
        glFlush()
        glPixelStorei(GL_PACK_ALIGNMENT, 4)
        data = glReadPixels(x1, y1, x1 + x2, y1 + y2, GL_RGBA, GL_UNSIGNED_BYTE)
        capture = np.frombuffer(data, dtype=np.uint8)
        capture = capture.reshape(y2 - y1, x2 - x1, 4)
        capture = np.flip(capture, axis=0)
        return capture

    def getPixelDepth(self, x: int, y: int) -> float:
        data = glReadPixels(x, self.screenHeight - 1 - y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT)
        zbuffer = float(data[0])
        depth = ZBuffer.zbuffer2depth(zbuffer)
        return depth

    def getPixelLayer(self, x: int, y: int) -> Tuple[Union[None, LayerGroup], int, Union[None, Layer], int]:
        depth = int(round(self.getPixelDepth(x, y)))
        for layerGroupIndex, layerGroup in enumerate(self.__layerGroups):
            if layerGroup is None:
                continue
            for layerIndex, layer in enumerate(layerGroup):
                if layer is None:
                    continue
                if layer.hasDepth(depth):
                    return layerGroup, layerGroupIndex, layer, layerIndex
        return None, -1, None, -1

    def quit(self):
        for layerGroup in self.__layerGroups:
            if layerGroup is None:
                continue
            layerGroup.deleteObjects()
        pygame.quit()
        quit()
