from __future__ import annotations

from abc import abstractmethod
from typing import Union, Tuple, List

import numpy as np

from gui.impl.opengl import OpenGLGUIFacade, ZBuffer
from gui.impl.opengl.layer import OpenGLLayerGroup
from gui.impl.opengl.private import Mesh, FaceAllocator


class OpenGLLayer:

    def __init__(self, gui: OpenGLGUIFacade, layerGroup: OpenGLLayerGroup, dynamic: bool):
        self._gui = gui
        self._layerGroup = layerGroup

        # Main mesh
        self.__mesh = Mesh(gui)
        self.__faceAllocator = FaceAllocator() if dynamic else None
        self.__needUpdate = True

        # Mesh cpu data
        self.__vertices = None  # type: Union[None, np.ndarray]
        self.__faces = None  # type: Union[None, np.ndarray]
        self.__uvMap = None  # type: Union[None, np.ndarray]

        # Transparent tile
        self.__transparentTile = (0, 0, 0, 0)  # type: Tuple[int, int, int, int]

        # Shader properties
        self.__uvShiftX = 0.0  # type: float
        self.__uvShiftY = 0.0  # type: float

    @abstractmethod
    def updateDepths(self, minDepth: int, maxDepth: int, index: int, maxIndex: int) -> int:
        raise NotImplementedError()

    @abstractmethod
    def hasDepth(self, depth: int) -> bool:
        raise NotImplementedError()

    def setTileset(self, fileName: Union[str, np.ndarray]):
        self.__mesh.setTileset(fileName)

    def setTileSize(self, tileWidth: int, tileHeight: int):
        self.__mesh.setTileSize(tileWidth, tileHeight)
        self.__transparentTile = (
            self.__mesh.textureWidth - self.__mesh.tileWidth,
            self.__mesh.textureHeight - self.__mesh.tileHeight,
            self.__mesh.textureWidth,
            self.__mesh.textureHeight
        )

    def deleteObjects(self):
        self.__mesh.deleteObjects()

    def setRenderedFaceCount(self, count: int):
        self.__mesh.setRenderedFaceCount(count)

    @property
    def tileWidth(self) -> int:
        return self.__mesh.tileWidth

    @property
    def tileHeight(self) -> int:
        return self.__mesh.tileHeight

    @property
    def texturePixelWidth(self) -> float:
        return self.__mesh.texturePixelWidth

    @property
    def texturePixelHeight(self) -> float:
        return self.__mesh.texturePixelHeight

    @property
    def faceCount(self) -> int:
        if self.__vertices is None:
            return 0
        return self.__vertices.shape[0]

    def createFaces(self, faceCount: int, depth: int) -> int:
        self.__vertices = None
        if self.__faceAllocator is not None:
            self.__faceAllocator.freeAll()
        return self.__addFaces(faceCount, depth)

    def allocateFaces(self, facesId: int, faceCount: int):
        assert self.__faceAllocator is not None, "Static layer, can't allocate/free faces"
        self.__faceAllocator.allocate(facesId, faceCount)
        if self.faceCount < self.__faceAllocator.faceCount:
            self.__setFaceCount(self.__faceAllocator.faceCount)
        self.setRenderedFaceCount(self.__faceAllocator.faceCount)
        return self.__faceAllocator.getIndices(facesId)

    def getAllocatedFaces(self, facesId: int):
        assert self.__faceAllocator is not None, "Static layer, can't allocate/free faces"
        return self.__faceAllocator.getIndices(facesId)

    def freeFaces(self, facesId: int):
        assert self.__faceAllocator is not None, "Static layer, can't allocate/free faces"
        faceIndices = self.__faceAllocator.getIndices(facesId)
        self.hideFaces(faceIndices)
        self.__faceAllocator.free(facesId)

    def freeAllFaces(self):
        assert self.__faceAllocator is not None, "Static layer, can't allocate/free faces"
        self.__faceAllocator.freeAll()
        self.__setFaceCount(0)

    def __addFaces(self, faceCount: int, depth: int) -> int:
        previousFaceCount = self.faceCount
        newFaceCount = previousFaceCount + faceCount

        if self.__vertices is None:
            self.__vertices = np.zeros([faceCount, 4, 3], dtype=np.float32)
            self.__uvMap = np.zeros([faceCount, 4, 2], dtype=np.float32)
            self.__faces = np.zeros([faceCount, 4], dtype=np.uint)
        else:
            self.__vertices.resize([newFaceCount, 4, 3])
            self.__uvMap.resize([newFaceCount, 4, 2])
            self.__faces.resize([newFaceCount, 4])

        for faceIndex in range(previousFaceCount, newFaceCount):
            self.setFacePixelLocation(faceIndex, 0, 0, self.__mesh.tileWidth, self.__mesh.tileHeight)
            self.setFacePixelTexture(
                faceIndex,
                self.__transparentTile[0], self.__transparentTile[1],
                self.__transparentTile[2], self.__transparentTile[3]
            )

            self.__faces[faceIndex, 0] = faceIndex * 4
            self.__faces[faceIndex, 1] = faceIndex * 4 + 1
            self.__faces[faceIndex, 2] = faceIndex * 4 + 2
            self.__faces[faceIndex, 3] = faceIndex * 4 + 3

        self.__vertices[previousFaceCount:newFaceCount, :, 2] = ZBuffer.depth2z(depth)
        self.__needUpdate = True

        return previousFaceCount

    def __removeFaces(self, firstIndex: int, lastIndex: int):
        assert self.__vertices is not None
        assert 0 <= firstIndex < self.__vertices.shape[0]
        assert 0 < lastIndex <= self.__vertices.shape[0]
        assert firstIndex < lastIndex

        # Delete all data related to the faces to remove
        self.__vertices = np.delete(self.__vertices, range(firstIndex, lastIndex), axis=0)
        self.__faces = np.delete(self.__faces, range(firstIndex, lastIndex), axis=0)
        self.__uvMap = np.delete(self.__uvMap, range(firstIndex, lastIndex), axis=0)

        # Update the face indexes
        faceCount = self.__faces.shape[0]
        for faceIndex in range(firstIndex, faceCount):
            self.__faces[faceIndex, 0] = faceIndex * 4
            self.__faces[faceIndex, 1] = faceIndex * 4 + 1
            self.__faces[faceIndex, 2] = faceIndex * 4 + 2
            self.__faces[faceIndex, 3] = faceIndex * 4 + 3

        self.__needUpdate = True

    def __setFaceCount(self, faceCount: int, depth: int = 0):
        diff = faceCount - self.faceCount
        if diff > 0:
            self.__addFaces(diff, depth)
        elif diff < 0:
            self.__removeFaces(faceCount, self.faceCount)

    def findFaces(self, x: float, y: float) -> List[int]:
        spriteScreenX = -1 + x * self.__mesh.screenPixelWidth
        spriteScreenY = 1 - y * self.__mesh.screenPixelHeight

        x1 = self.__vertices[:, 1, 0]
        y1 = self.__vertices[:, 1, 1]
        x2 = self.__vertices[:, 3, 0]
        y2 = self.__vertices[:, 3, 1]

        mask = (x1 <= spriteScreenX <= x2) and (y2 <= spriteScreenY <= y1)
        return mask.nonzero()[0].tolist()

    def getFacePixelLocation(self, faceIndex) -> (float, float, float, float):
        assert 0 <= faceIndex < self.faceCount

        spriteScreenX1, spriteScreenY1 = self.__vertices[faceIndex, 1, 0:2]
        spriteScreenX2, spriteScreenY2 = self.__vertices[faceIndex, 3, 0:2]
        x1 = (spriteScreenX1 + 1) / self.__mesh.screenPixelWidth
        y1 = -(spriteScreenY1 - 1) / self.__mesh.screenPixelHeight
        x2 = (spriteScreenX2 + 1) / self.__mesh.screenPixelWidth
        y2 = -(spriteScreenY2 - 1) / self.__mesh.screenPixelHeight
        return x1, y1, x2, y2

    def setFacePixelLocation(self, faceIndex: int, x1: float, y1: float, x2: float, y2: float):
        assert 0 <= faceIndex < self.faceCount

        spriteScreenX1 = -1 + x1 * self.__mesh.screenPixelWidth
        spriteScreenY1 = 1 - y1 * self.__mesh.screenPixelHeight
        spriteScreenX2 = -1 + x2 * self.__mesh.screenPixelWidth
        spriteScreenY2 = 1 - y2 * self.__mesh.screenPixelHeight
        self.__vertices[faceIndex, 0, 0:2] = [spriteScreenX1, spriteScreenY2]
        self.__vertices[faceIndex, 1, 0:2] = [spriteScreenX1, spriteScreenY1]
        self.__vertices[faceIndex, 2, 0:2] = [spriteScreenX2, spriteScreenY1]
        self.__vertices[faceIndex, 3, 0:2] = [spriteScreenX2, spriteScreenY2]

        self.__needUpdate = True

    def shiftFacesPixelLocation(self, faceIndices: List[int], dx: float, dy: float):
        dx = dx * self.__mesh.screenPixelWidth
        dy = -dy * self.__mesh.screenPixelHeight
        shift = np.array([dx, dy], dtype=np.float)
        self.__vertices[faceIndices, :, 0:2] += shift
        self.__needUpdate = True

    def setFacePixelTexture(self, faceIndex: int, tileX1: float, tileY1: float, tileX2: float, tileY2: float):
        assert 0 <= faceIndex < self.faceCount

        spriteTextureX1 = tileX1 / self.__mesh.textureWidth
        spriteTextureY1 = tileY1 / self.__mesh.textureHeight
        spriteTextureX2 = tileX2 / self.__mesh.textureWidth
        spriteTextureY2 = tileY2 / self.__mesh.textureHeight
        self.__uvMap[faceIndex, 0] = [spriteTextureX1, spriteTextureY2]
        self.__uvMap[faceIndex, 1] = [spriteTextureX1, spriteTextureY1]
        self.__uvMap[faceIndex, 2] = [spriteTextureX2, spriteTextureY1]
        self.__uvMap[faceIndex, 3] = [spriteTextureX2, spriteTextureY2]

        self.__needUpdate = True

    def setFaceTileLocation(self, faceIndex: int, x: float, y: float, width: float = 1, height: float = 1):
        assert 0 <= faceIndex < self.__vertices.shape[0]

        spriteScreenX1 = -1 + (x + 1 - width) * self.__mesh.screenTileWidth
        spriteScreenY1 = 1 - (y + 1 - height) * self.__mesh.screenTileHeight
        spriteScreenX2 = spriteScreenX1 + self.__mesh.screenTileWidth * width
        spriteScreenY2 = spriteScreenY1 - self.__mesh.screenTileHeight * height
        self.__vertices[faceIndex, 0, 0:2] = [spriteScreenX1, spriteScreenY2]
        self.__vertices[faceIndex, 1, 0:2] = [spriteScreenX1, spriteScreenY1]
        self.__vertices[faceIndex, 2, 0:2] = [spriteScreenX2, spriteScreenY1]
        self.__vertices[faceIndex, 3, 0:2] = [spriteScreenX2, spriteScreenY2]

        self.__needUpdate = True

    def setFaceDepth(self, faceIndex: int, depth: float):
        assert 0 <= faceIndex < self.__vertices.shape[0]
        self.__vertices[faceIndex, :, 2] = ZBuffer.depth2z(depth)
        self.__needUpdate = True

    def setFacesDepth(self, depth: float, faceIndices: Union[None, range] = None):
        if self.__vertices is None:
            return
        if faceIndices is None:
            self.__vertices[..., 2] = ZBuffer.depth2z(depth)
        else:
            self.__vertices[faceIndices, :, 2] = ZBuffer.depth2z(depth)
        self.__needUpdate = True

    def setFaceTileTexture(self, faceIndex: int, tileX: float, tileY: float, width: float = 1, height: float = 1):
        assert 0 <= faceIndex < self.__vertices.shape[0]

        spriteTextureX1 = (tileX + 1 - width) * self.__mesh.textureTileWidth
        spriteTextureY1 = (tileY + 1 - height) * self.__mesh.textureTileHeight
        spriteTextureX2 = spriteTextureX1 + self.__mesh.textureTileWidth * width
        spriteTextureY2 = spriteTextureY1 + self.__mesh.textureTileHeight * height
        self.__uvMap[faceIndex, 0] = [spriteTextureX1, spriteTextureY2]
        self.__uvMap[faceIndex, 1] = [spriteTextureX1, spriteTextureY1]
        self.__uvMap[faceIndex, 2] = [spriteTextureX2, spriteTextureY1]
        self.__uvMap[faceIndex, 3] = [spriteTextureX2, spriteTextureY2]

        self.__needUpdate = True

    def hideFace(self, faceIndex: int):
        self.setFacePixelTexture(
            faceIndex,
            self.__transparentTile[0], self.__transparentTile[1],
            self.__transparentTile[2], self.__transparentTile[3]
        )

    def hideFaces(self, faceIndices: List[int]):
        for faceIndex in faceIndices:
            self.setFacePixelTexture(
                faceIndex,
                self.__transparentTile[0], self.__transparentTile[1],
                self.__transparentTile[2], self.__transparentTile[3]
            )

    @property
    def uvShiftX(self) -> float:
        return self.__uvShiftX

    @property
    def uvShiftY(self) -> float:
        return self.__uvShiftY

    def setTilesetShift(self, x: float, y: float):
        self.__uvShiftX = -x * self.texturePixelWidth
        self.__uvShiftY = y * self.texturePixelHeight

    def draw(self):
        if self.__needUpdate and self.__vertices is not None:
            self.__mesh.updateData(self.__vertices, self.__faces, self.__uvMap)
            self.__needUpdate = False
        self.__mesh.draw()
