from __future__ import annotations

import typing
from typing import Union, Dict, List

from gui.impl.opengl import OpenGLGUIFacade
from gui.impl.pygame.text import TextTileset
from gui.layer import TextLayer
from gui.text import TextStyle, TextToken, TextTokenType, ContentTextToken, StyleTextToken, TextTokenizer
from .OpenGLLayer import OpenGLLayer
from .OpenGLLayerGroup import OpenGLLayerGroup
from ..private import FaceAllocator, TextBox


class OpenGLTextLayer(OpenGLLayer, TextLayer):

    def __init__(self, gui: OpenGLGUIFacade, layerGroup: OpenGLLayerGroup):
        super(OpenGLTextLayer, self).__init__(gui, layerGroup, dynamic=True)
        self.__tileset = TextTileset()
        self.__lastId = 0
        self.__baseStyle = None  # type: Union[None, TextStyle]
        self.__depth = 0  # type: int
        self.__maxCharacterCount = 10000
        self.__textBoxes = {}  # type: Dict[int, TextBox]

    def updateDepths(self, minDepth: int, maxDepth: int, index: int, maxIndex: int) -> int:
        self.__depth = minDepth
        self.setFacesDepth(self.__depth)
        return self.__depth - 1

    def hasDepth(self, depth: int) -> bool:
        return depth == self.__depth

    def setStyle(self, style: TextStyle):
        self.__baseStyle = style

    def createTextBox(self, x: int = 0, y: int = 0) -> int:
        self.__lastId += 1
        textBoxId = self.__lastId
        self.__textBoxes[textBoxId] = TextBox(x, y)
        return textBoxId

    def setTextBoxContent(self, textBoxId: int, content: str):
        textBox = self.__textBoxes[textBoxId]
        if textBox.content == content:
            return
        textBox.content = content
        tokenizer = TextTokenizer(self.__baseStyle)
        tokens = tokenizer.parse(content)
        self.__setTextBoxTokens(textBoxId, textBox.x1, textBox.y1, tokens)

    def setTextBoxLocation(self, textBoxId: int, x: float, y: float):
        self.__setTextBoxLocation(textBoxId, x, y)

    def getTextBoxSize(self, textBoxId: int) -> (int, int):
        textBox = self.__textBoxes[textBoxId]
        return textBox.x2 - textBox.x1, textBox.y2 - textBox.y1

    def deleteTextBox(self, textBoxId: int):
        self.freeFaces(textBoxId)

    def deleteTextBoxes(self):
        self.freeAllFaces()

    ###############################################################################################

    def __updateTileset(self, characters: str):
        # Add characters to the tileset
        if self.__tileset.addCharacters(characters):
            # Send the tileset to the GPU
            self.setTileset(self.__tileset.toNumpyArray())
            if False:  # Save generated tileset (to debug/illustrate)
                from PIL import Image
                if not hasattr(self, "tilesetUpdates"):
                    self.tilesetUpdates = 0
                self.tilesetUpdates += 1
                Image.fromarray(self.__tileset.toNumpyArray()).save("charactersTileset{}.png".format(self.tilesetUpdates))

    def __setTextBoxTokens(self, textBoxId: int, x: float, y: float, tokens: List[TextToken]):
        # First make sure that all character are in the tileset
        characterCount = 0
        for token in tokens:
            if token.type == TextTokenType.CONTENT:
                contentTextToken = typing.cast(ContentTextToken, token)
                content = contentTextToken.content
                self.__updateTileset(content)
                characterCount += len(content)
            elif token.type == TextTokenType.STYLE:
                styleTextToken = typing.cast(StyleTextToken, token)
                self.__tileset.setStyle(styleTextToken.style)
            else:
                raise RuntimeError("Unsupported token type")

        # Allocate faces
        self.freeFaces(textBoxId)
        maxCharacterCount = min(characterCount, self.__maxCharacterCount)
        faceIndices = self.allocateFaces(textBoxId, maxCharacterCount)
        self.setFacesDepth(self.__depth, faceIndices)

        # Set the faces
        x0 = x
        charIndex = 0
        textBox = self.__textBoxes[textBoxId]
        textBox.x1 = x
        textBox.y1 = y
        textBox.x2 = x
        textBox.y2 = y
        for token in tokens:
            if token.type == TextTokenType.CONTENT:
                contentTextToken = typing.cast(ContentTextToken, token)
                for character in contentTextToken.content:
                    if character == "\n":
                        x = x0
                        style = self.__tileset.getStyle()
                        y += style.tileHeight
                        continue
                    tile = self.__tileset.getCharacterTile(character)
                    self.setFacePixelTexture(faceIndices[charIndex], tile.x1, tile.y1, tile.x2, tile.y2)
                    self.setFacePixelLocation(faceIndices[charIndex], x, y, x + tile.width, y + tile.height)
                    textBox.x2 = max(textBox.x2, x + tile.width)
                    textBox.y2 = max(textBox.y2, y + tile.height)
                    x += tile.advance
                    charIndex += 1
                    if charIndex >= maxCharacterCount:
                        break
            elif token.type == TextTokenType.STYLE:
                styleTextToken = typing.cast(StyleTextToken, token)
                self.__tileset.setStyle(styleTextToken.style)
            else:
                raise RuntimeError("Unsupported token type")
            if charIndex >= maxCharacterCount:
                break

    def __setTextBoxLocation(self, textBoxId: int, x: float, y: float):
        faceIndices = self.getAllocatedFaces(textBoxId)
        if len(faceIndices) == 0:
            return
        x1, y1, x2, y2 = self.getFacePixelLocation(faceIndices[0])
        dx = x - x1
        dy = y - y1
        self.shiftFacesPixelLocation(faceIndices, dx, dy)
        textBox = self.__textBoxes[textBoxId]
        textBox.x1 += dx
        textBox.y1 += dy
        textBox.x2 += dx
        textBox.y2 += dy
