from typing import Union, Dict

from gui import GUIFacade
from gui.layer import CharactersLayer, LayerGroup, TextLayer
from gui.text import TextStyle
from state import Region, Character
from .Renderer import Renderer
from .animation import CharacterAnimation


class CharactersRenderer(Renderer):

    def __init__(self, gui: GUIFacade, layerGroup: LayerGroup, layerIndex: int, textLayerIndex: int, tileSize: int):
        super(CharactersRenderer, self).__init__(gui)
        self.__layerGroup = layerGroup
        self.__layerIndex = layerIndex
        self.__tileSize = tileSize  # type: int
        self.__layer = None  # type: Union[None, CharactersLayer]
        self.__animations = {}  # type: Dict[int, CharacterAnimation]

        # Text
        self.__textLayer = None  # type: Union[None, TextLayer]
        self.__textLayerIndex = textLayerIndex
        self.__textStyle = TextStyle("assets/font/basis33/basis33.ttf", self.__tileSize, (255, 255, 255))
        self.__textStyle.outline = True
        self.__textStyle.outlineColor = (0, 0, 0)
        self.__textStyle.outlineThickness = 2

    def updateAll(self, region: Union[None, Region]):

        if self.__layer is not None:
            self.__layerGroup.removeLayer(self.__layer)
            self.__layer = None
        if self.__textLayer is not None:
            self.__layerGroup.removeLayer(self.__textLayer)
            self.__textLayer = None

        # Case where we remove the region
        if region is None:
            return

        characters = region.characters

        # Create the layer
        self.__layer = self.__layerGroup.createCharactersLayer(self.__layerIndex)
        self.__layer.setTileset("assets/tile/characters.png")
        self.__layer.setTileSize(self.__tileSize, self.__tileSize)
        self.__layer.setGridSize(region.width, region.height)
        self.__layer.setCharacterCount(len(characters))
        self.__animations = {}

        # Text layer
        self.__textLayer = self.__layerGroup.createTextLayer(self.__textLayerIndex)
        self.__textLayer.setStyle(self.__textStyle)

        for index in range(len(characters)):
            self.updateCharacter(index, characters[index])

    def updateCharacter(self, index: int, character: Character):
        if index >= self.__layer.getCharacterCount():
            self.__layer.setCharacterCount(index + 1)

        if character is not None:
            if index not in self.__animations:
                self.__animations[index] = CharacterAnimation(self.__layer, index, self.__textLayer)
            self.__animations[index].updateCharacter(character, self._gui.getTime())
        else:
            if index in self.__animations:
                self.__animations[index].delete()
                del self.__animations[index]

    def updateAnimations(self, time: float, epochFrame: float):
        toDelete = []
        for index, animation in self.__animations.items():
            if not animation.update(time, epochFrame):
                toDelete.append(index)
        for index in toDelete:
            self.__animations[index].delete()
            del self.__animations[index]

    def getCharacterLocation(self, characterIndex: int) -> (float, float):
        if characterIndex in self.__animations:
            return self.__animations[characterIndex].getLocation()
        return -1, -1

