from typing import Union

import numpy as np

from gui import GUIFacade
from gui.layer import GridLayer, LayerGroup
from state import Region
from .Renderer import Renderer


class GroundRenderer(Renderer):

    def __init__(self, gui: GUIFacade, layerGroup: LayerGroup, layerIndex: int, tileSize: int):
        super(GroundRenderer, self).__init__(gui)
        self.__layerGroup = layerGroup
        self.__layerIndex = layerIndex
        self.__tileSize = tileSize
        self.__layer = None  # type: Union[None, GridLayer]
        self.__regionName = None  # type: Union[None, str]
        self.__region = None  # type: Union[None, Region]

    def updateAll(self, region: Union[None, Region]):
        # Case where we remove the region
        if region is None:
            if self.__layer is not None:
                self.__layerGroup.removeLayer(self.__layer)
                self.__layer = None
            return

        # Region size
        width = region.width
        height = region.height

        # Create the layer
        self.__layer = self.__layerGroup.createGridLayer(self.__layerIndex)
        self.__layer.setTileset("assets/tile/groundcombi.png")
        self.__layer.setTileSize(self.__tileSize, self.__tileSize)
        self.__layer.setTiles(np.zeros([width, height, 2], dtype=np.int32))

        # Update tiles
        for y in range(height):
            for x in range(width):
                self.updateTile(region, x, y)

    def updateTile(self, region: Region, x: int, y: int):
        neightbors = region.getGroundNeightbors(x, y)
        value = neightbors[1, 1]
        otherValue = min(value, neightbors[1, 0], neightbors[1, 2], neightbors[0, 1], neightbors[2, 1])
        if otherValue == value:
            otherValue = neightbors.min()

        tilesetLocation = self.groundTilesetLocations[value, otherValue]
        tileX = tilesetLocation[0] * 8
        tileY = tilesetLocation[1] * 6
        if value == otherValue:
            tileX += 6
            tileY += 1
        else:
            mask = neightbors != otherValue
            if mask[0, 0] and (not mask[0, 1] or not mask[1, 0]):
                mask[0, 0] = 0
            if mask[0, 2] and (not mask[0, 1] or not mask[1, 2]):
                mask[0, 2] = 0
            if mask[2, 0] and (not mask[1, 0] or not mask[2, 1]):
                mask[2, 0] = 0
            if mask[2, 2] and (not mask[1, 2] or not mask[2, 1]):
                mask[2, 2] = 0
            powers = np.array([1, 2, 4, 8, 0, 16, 32, 64, 128], dtype=np.int32)
            code = (mask.flatten() * powers).sum()

            shift = self.code2tile[code]
            tileX += shift[0]
            tileY += shift[1]

        self.__layer.setTile(x, y, tileX, tileY)

    groundTilesetLocations = np.array([
        [[7, 7], [7, 7], [7, 7], [7, 7], [7, 7], [7, 7]],  # Invalid
        [[7, 7], [0, 0], [0, 0], [1, 0], [2, 0], [3, 0]],  # Grass
        [[7, 7], [0, 1], [0, 1], [1, 1], [2, 1], [3, 1]],  # Swamp
        [[7, 7], [0, 2], [1, 2], [0, 2], [2, 2], [3, 2]],  # Dirt
        [[7, 7], [0, 3], [1, 3], [2, 3], [0, 3], [3, 3]],  # Sand
        [[7, 7], [0, 4], [1, 4], [2, 4], [3, 4], [0, 4]],  # Dead land
    ], dtype=np.int32)

