from typing import Union

import numpy as np

from gui import GUIFacade
from gui.layer import GridLayer, LayerGroup
from state import Region
from .Renderer import Renderer


class WaterRenderer(Renderer):

    def __init__(self, gui: GUIFacade, layerGroup: LayerGroup, layerIndex: int, tileSize: int):
        super(WaterRenderer, self).__init__(gui)
        self.__layerGroup = layerGroup
        self.__layerIndex = layerIndex
        self.__tileSize = tileSize  # type: int
        self.__layer = None  # type: Union[None, GridLayer]
        self._animationStep = 0

    def updateAll(self, region: Union[None, Region]):
        # Case where we remove the region
        if region is None:
            if self.__layer is not None:
                self.__layerGroup.removeLayer(self.__layer)
                self.__layer = None
            return

        # Region size
        width = region.width
        height = region.height

        # Create the layer
        self.__layer = self.__layerGroup.createGridLayer(self.__layerIndex)
        self.__layer.setTileset("assets/tile/water.png")
        self.__layer.setTileSize(self.__tileSize, self.__tileSize)
        self.__layer.setTiles(np.zeros([width, height, 2], dtype=np.int32))

        # Update tiles
        for y in range(height):
            for x in range(width):
                self.updateTile(region, x, y)

    def updateTile(self, region: Region, x: int, y: int):
        if region.getWaterValue(x, y) == 0:
            self.__layer.setTile(x, y, 63, 63)
            return

        neightbors = region.getWaterNeightbors(x, y)
        value = neightbors[1, 1]
        tilesetLocation = self.waterTilesetLocations[value]
        tileX = tilesetLocation[0] * 8
        tileY = tilesetLocation[1] * 6

        mask = neightbors != 0
        if mask[0, 0] and (not mask[0, 1] or not mask[1, 0]):
            mask[0, 0] = 0
        if mask[0, 2] and (not mask[0, 1] or not mask[1, 2]):
            mask[0, 2] = 0
        if mask[2, 0] and (not mask[1, 0] or not mask[2, 1]):
            mask[2, 0] = 0
        if mask[2, 2] and (not mask[1, 2] or not mask[2, 1]):
            mask[2, 2] = 0
        powers = np.array([1, 2, 4, 8, 0, 16, 32, 64, 128], dtype=np.int32)
        code = (mask.flatten() * powers).sum()

        shift = self.code2tile[code]
        tileX += shift[0]
        tileY += shift[1]

        self.__layer.setTile(x, y, tileX, tileY)

    waterTilesetLocations = np.array([
        [7, 7], [0, 0], [1, 1], [2, 2], [3, 3], [4, 4]
    ], dtype=np.int32)

    def updateAnimations(self, time: float, epochFrame: float):
        if self.__layer is None:
            return
        self.__layer.setTilesetShift(
            self.__tileSize * 8 * (self._animationStep // 6), 0
        )
        self._animationStep += 1
        if self._animationStep >= 8 * 6:
            self._animationStep = 0


